#!/usr/bin/env python2
import os
import sys
import time
import glob
import shutil
import urlparse
import urllib2
import pwd

from charmhelpers.core.hookenv import (
    log,
    Hooks,
    config,
    relation_get,
    related_units,
    relation_ids,
    relation_set,
)

from charmhelpers.core.host import (
    rsync,
    mkdir,
)

from charmhelpers.fetch import (
    install_remote,
)

hooks = Hooks()
hook_dir = os.path.abspath(os.path.dirname(__file__))
charm_dir = os.path.dirname(hook_dir)
template_dir = os.path.join(charm_dir, 'templates')
fetch_dir = os.path.join(charm_dir, 'fetched')
revno_file = '/var/lib/juju/content-revno'


@hooks.hook('install')
def install():
    log("No action needed")


def prune_dirs(dir_glob, keep_dirs):
    """Get a list of the content directories and remove the oldest until there
    are only keep_dirs left"""
    direcs = []
    for direc in glob.glob(dir_glob):
        direcs.append(direc)
    direcs.sort()
    direcs.reverse()
    to_delete = direcs[keep_dirs:]
    for direc in to_delete:
        shutil.rmtree(direc)


def get_current_revno():
    if not os.path.isfile(revno_file):
        return -1
    with open(revno_file, 'r') as f:
        revno = f.read()
    if revno == '':
        return -1
    return int(revno)


def set_revno(revno):
    with open(revno_file, 'w') as content_file:
        content_file.write(str(revno))


def deploy_private_key(src_key, target_key):
    with open(src_key, 'r') as content_file:
        private_key = content_file.read()
    with open(target_key, 'w') as content_file:
        os.chmod(target_key, 0600)
        content_file.write(private_key)


def add_ssh_hostconfig(src_server, installed_key, ssh_config,
                       fetcher_host_key_file, remote_host_key=None):
    ssh_config_tmpl = os.path.join(template_dir, 'ssh_config.tmpl')
    with open(ssh_config_tmpl, 'r') as f:
        template = f.read()
    template_args = {'src_server': src_server, 'key_file': installed_key}
    open_mode = 'w'
    # If the ssh config files exists then check if it has an entry for the src
    # server
    if os.path.isfile(ssh_config):
        for line in open(ssh_config):
            if "Host " + src_server in line:
                return
        open_mode = 'a'
    # If a hostkey has been provided check against the file containing it
    # otherwise don't verify hostkeys
    with open(ssh_config, open_mode) as f:
        os.chmod(ssh_config, 0644)
        f.write(template.format(**template_args))
        if remote_host_key:
            f.write('   UserKnownHostsFile ' + fetcher_host_key_file + '\n')
        else:
            f.write('   StrictHostKeyChecking no' + '\n')


def add_ssh_hostkey(host_key_file, host_key):
    with open(host_key_file, 'w') as f:
        os.chmod(host_key_file, 0644)
        f.write(host_key + '\n')


def setup_ssh(src_server):
    euser = pwd.getpwuid(os.geteuid())[0]
    euser_home = os.path.expanduser("~" + euser)
    ssh_dir = os.path.join(euser_home, '.ssh')
    ssh_config = os.path.join(ssh_dir, 'config')
    fetcher_host_key_file = os.path.join(ssh_dir, 'fetcher_host_key')
    charm_key_dir = os.path.join(charm_dir, 'ssh-key')
    charm_key = os.path.join(charm_key_dir, 'content_fetcher_rsa')
    installed_key = os.path.join(ssh_dir, 'content_fetcher_rsa')
    remote_host_key = config().get('ssh_host_key')
    # Create .ssh in homedir if needed
    if not os.path.exists(ssh_dir):
        os.mkdir(ssh_dir, mode=0700)
    # Copy in the private key from the charm
    if os.path.isfile(charm_key):
        deploy_private_key(charm_key, installed_key)
    # If a hostkey has been provided write it out to its own file to be
    # referenced by the users ssh config later
    if len(remote_host_key) > 0:
        add_ssh_hostkey(fetcher_host_key_file, remote_host_key)
    add_ssh_hostconfig(src_server, installed_key, ssh_config,
                       fetcher_host_key_file, remote_host_key=remote_host_key)


@hooks.hook('upgrade-charm')
@hooks.hook('remote-content-relation-changed')
@hooks.hook('config-changed')
def config_changed():
    dst_dir = config().get('dest_dir')
    trigger_revno = config().get('deploy_trigger')
    current_revno = get_current_revno()
    auto_update = config().get('update_on_change')
    if auto_update or trigger_revno > current_revno:
        rid = relation_ids(reltype='remote-content')
        if rid:
            unit = related_units(relid=rid[0])[0]
            requested_dst_dir = relation_get(attribute='dest_dir', unit=unit,
                                             rid=rid[0])
            if requested_dst_dir:
                dst_dir = requested_dst_dir
        deploy_archive(dst_dir)
        if rid:
            relation_set(relation_id=rid[0], revno=trigger_revno)
    else:
        log("The deploy_trigger matches previous deploy.")
        log("Not updating content.")


def create_deploy_dir(dst_dir):
    if os.path.exists(dst_dir):
        backup_postfix = time.strftime("-%Y%m%d-%H%M%S", time.gmtime())
        backup_dir = dst_dir.rstrip('/') + backup_postfix
        log("Backing up old content to " + backup_dir)
        mkdir(backup_dir, perms=0755)
        rsync(from_path=dst_dir + '/', to_path=backup_dir, options=['--delete', '--executability', '-l'])
    else:
        mkdir(dst_dir, perms=0755)


def wipe_fetch_area():
    # Clean out the tmp area that may hold an old branch
    for direc in glob.glob(os.path.join(fetch_dir, '*')):
        if os.path.isdir(direc):
            shutil.rmtree(direc)


def deploy_archive(dst_dir):
    trigger_revno = config().get('deploy_trigger')
    arch_url = config().get('archive_location')
    if len(arch_url) == 0:
        log("No archive location specified")
        return
    url_parts = urlparse.urlparse(arch_url)
    src_type = url_parts[0]
    wipe_fetch_area()
    if len(src_type) == 0 and len(config().get('tenant_name')) > 0:
        import swift
        sc = swift.SwiftFetchHandler(config().get('auth_url'),
                                     config().get('username'),
                                     config().get('secret'),
                                     config().get('tenant_name'))
        archive_dir = sc.install(config().get('container_name'),
                                 arch_url)
    elif src_type == 'bzr+ssh':
        url_parts = urlparse.urlparse(arch_url)
        auth, src_server = urllib2.splituser(url_parts.netloc)
        setup_ssh(src_server)
        archive_dir = install_remote(arch_url)
    else:
        archive_dir = install_remote(arch_url)
    create_deploy_dir(dst_dir)
    log("Installing new content")
    rsync(from_path=archive_dir + '/', to_path=dst_dir, options=['--delete', '--executability', '-l'])
    backup_dir_glob = dst_dir.rstrip('/') + "-*"
    prune_dirs(backup_dir_glob, config().get('keep_dir_count'))
    set_revno(trigger_revno)


if __name__ == "__main__":
    hooks.execute(sys.argv)
