#!/usr/bin/python
import argparse
import netifaces
import os
import pwd
import random
import shutil
import string
import sys
import tarfile
import tempfile
import time

from os.path import dirname
from zope.interface import implements
from twisted.internet import reactor, ssl
from twisted.web import server, resource, guard, static
from twisted.cred.portal import IRealm, Portal
from twisted.protocols.ftp import FTPFactory, FTPRealm
from twisted.cred.checkers import AllowAnonymousAccess

from twisted.cred.checkers import \
    InMemoryUsernamePasswordDatabaseDontUse as InMemoryDB
from Crypto.PublicKey import RSA


def get_listen_ip():
    for i in netifaces.interfaces():
        if i == 'lo':
            continue
        return netifaces.ifaddresses(i).get(netifaces.AF_INET)[0]['addr']


def create_work_area():
    return tempfile.mkdtemp()


def create_test_html(pub_dir, access_type):
    test_file = os.path.join(pub_dir, 'index.html')
    f = open(test_file, 'w')
    f.write('<html><body><h1>Website collected from %s at %s</h1>\
             </body></html>\n' %
            (access_type, time.strftime("%Y/%m/%d %H:%M:%S", time.gmtime())))
    f.close()
    return test_file


def create_test_tarball(pub_dir, access_type):
    tar_file = os.path.join(pub_dir, 'content.tar')
    test_file = create_test_html(pub_dir, access_type)
    tar = tarfile.open(tar_file, "w")
    tar.add(test_file, arcname="index.html")
    tar.close()
    os.remove(test_file)


def create_test_bzrrepo(pub_dir, access_type):
    from bzrlib.bzrdir import BzrDir
    branch = BzrDir.create_branch_convenience(work_area, force_new_tree=True)
    test_file = create_test_html(work_area, 'bzr+ssh')
    branch.get_config().set_user_option('email', u'Branch Identity' +
                                        'content-fetcher@%s' % get_listen_ip())
    wtree = branch.bzrdir.open_workingtree()
    wtree.smart_add([test_file])
    wtree.commit('Committing test data')


def tidyup(work_area):
    shutil.rmtree(work_area)


def get_charm_root():
    return dirname(dirname(os.path.realpath(__file__)))


def display_juju_set_cmd(src_uri=None):
    if src_uri:
        print "juju set content-fetcher archive_location=%s" % (src_uri)
    print "juju set content-fetcher deploy_trigger=$(juju get" + \
          " content-fetcher | grep -A6 deploy_trigger | " + \
          " awk '/value:/ {print $2 + 1}')"
    sys.stdout.flush()


def write_sshkey(key, key_type, key_file, write_mode='w'):
    with open(key_file, write_mode) as content_file:
        os.chmod(key_file, 0600)
        content_file.write(key.exportKey(key_type) + '\n')


class setup_test_listener:
    def __init__(self, work_area, port, proto, **users):
        self.work_area = work_area
        self.port = port
        self.proto = proto
        self.users = users
        if users:
            self.auth_type = 'auth'
        else:
            self.auth_type = 'anon'
        self.create_payload()

    def create_payload(self):
        create_test_tarball(work_area, '%s %s' % (self.proto, self.auth_type))
        if self.users:
            for user in self.users:
                os.mkdir(os.path.join(work_area, user))
                create_test_tarball(os.path.join(work_area, user), '%s %s' %
                                    (self.proto, self.auth_type))

    def display_access_info(self):
        if self.users:
            for user in self.users:
                display_juju_set_cmd("%s://%s:%s@%s:%s/content.tar" %
                                     (self.proto, user, self.users[user],
                                      get_listen_ip(), self.port))
        else:
            display_juju_set_cmd("%s://%s:%s/content.tar" %
                                 (self.proto, get_listen_ip(), self.port))

    def start_server(self):
        if self.proto in ['http', 'https']:
            self.start_http_server()
        elif self.proto == 'ftp':
            self.start_ftp_server()

    def start_http_server(self):
        class SimpleRealm(object):
            implements(IRealm)

            def requestAvatar(self, avatarId, mind, *interfaces):
                if resource.IResource in interfaces:
                    return resource.IResource, root, lambda: None
                raise NotImplementedError()

        if self.users:
            checkers = [InMemoryDB(**self.users)]
        else:
            checkers = [AllowAnonymousAccess()]
        wrapper = guard.HTTPAuthSessionWrapper(
            Portal(SimpleRealm(), checkers),
            [guard.BasicCredentialFactory('test realm')])
        if self.proto == 'https':
            sslContext = ssl.DefaultOpenSSLContextFactory(
                os.path.join(get_charm_root(), 'testserver',
                             'website-key.pem'),
                os.path.join(get_charm_root(), 'testserver',
                             'website-cert.pem'),
            )
            reactor.listenSSL(self.port, server.Site(resource=wrapper),
                              contextFactory=sslContext)
        else:
            reactor.listenTCP(self.port, server.Site(resource=wrapper))
        try:
            reactor.run()
        finally:
            tidyup(self.work_area)

    def start_ftp_server(self):
        p = Portal(FTPRealm(self.work_area, userHome=self.work_area),
                   (AllowAnonymousAccess(),
                    InMemoryDB(**self.users),
                    )
                   )
        f = FTPFactory(p)
        try:
            reactor.listenTCP(self.port, f)
            reactor.run()
        finally:
            tidyup(self.work_area)


class setup_test_files:
    def __init__(self, work_area, proto, swift_creds={}, add_pub_key=None,
                 **users):
        self.work_area = work_area
        self.proto = proto
        self.users = users
        self.swift_creds = swift_creds
        self.add_pub_key = add_pub_key
        if users:
            self.auth_type = 'auth'
        else:
            self.auth_type = 'anon'
        self.create_payload()

    def create_payload(self):
        if self.proto == 'bzr+ssh':
            create_test_bzrrepo(self.work_area, self.proto)
        else:
            create_test_tarball(self.work_area, self.proto)

    def deploy_payload(self):
        if self.proto == 'swift':
            self.deploy_swift_payload()
        elif self.proto == 'bzr+ssh':
            self.deploy_bzrssh_payload()
        elif self.proto == 'file':
            self.deploy_file_payload()

    def set_swift_creds(self, auth_url, username, secret, tenant_name,
                        archive_location, container_name):
        self.swift_creds['auth_url'] = auth_url
        self.swift_creds['username'] = username
        self.swift_creds['secret'] = secret
        self.swift_creds['tenant_name'] = tenant_name
        self.swift_creds['archive_location'] = archive_location
        self.swift_creds['container_name'] = container_name

    def deploy_swift_payload(self):
        import swift
        sh = swift.SwiftFetchHandler(self.swift_creds['auth_url'],
                                     self.swift_creds['username'],
                                     self.swift_creds['secret'],
                                     self.swift_creds['tenant_name'])
        sc = sh.get_connection()
        with open(os.path.join(work_area, 'content.tar'), 'r') as f:
            fcontent = f.read()
        sc.put_container(self.swift_creds['container_name'])
        sc.put_object(self.swift_creds['container_name'], 'content.tar',
                      fcontent)
        cf_params = ' '.join(["%s=%s" % (k, v)
                              for k, v in self.swift_creds.iteritems()])
        print "juju set content-fetcher " + cf_params
        display_juju_set_cmd()
        tidyup(work_area)

    def deploy_bzrssh_payload(self):
        charm_key_dir = os.path.join(get_charm_root(), 'ssh-key')
        charm_key_file = os.path.join(charm_key_dir, 'content_fetcher_rsa')
        euser = pwd.getpwuid(os.geteuid())[0]
        euser_home = os.path.expanduser("~" + euser)
        ssh_dir = os.path.join(euser_home, '.ssh')
        local_privkey_file = os.path.join(ssh_dir, 'content_fetcher_rsa')
        local_authkey_file = os.path.join(ssh_dir, 'authorized_keys')
        print "To test this transport please:"
        if os.path.isfile(charm_key_file):
            with open(charm_key_file, 'r') as content_file:
                fcontent = content_file.read()
            key = RSA.importKey(fcontent)
        else:
            key = RSA.generate(2048)
            if os.path.isdir(charm_key_dir):
                write_sshkey(key, 'PEM', charm_key_file)
            else:
                write_sshkey(key, 'PEM', local_privkey_file)
                print "0) Copy key from %s to %s on the deployment server" % \
                    (local_privkey_file, local_privkey_file)
        pubkey = key.publickey()

        print "* juju upgrade-charm content-fetcher"
        if self.add_pub_key:
            print local_authkey_file
            write_sshkey(pubkey, 'OpenSSH', local_authkey_file, write_mode='a')
        else:
            print "* Add public key below to authorized_keys file of the user"
            print "  used for the content fetch"
            print pubkey.exportKey('OpenSSH')
        display_juju_set_cmd('bzr+ssh://%s@%s/%s' % (os.environ['USER'],
                                                     get_listen_ip(),
                                                     work_area))

    def deploy_file_payload(self):
        shutil.copy(work_area + '/content.tar', '/tmp/content.tar')
        print "juju scp /tmp/content.tar content-fetcher/0:/tmp/"
        display_juju_set_cmd('file:///tmp/content.tar')


if __name__ == '__main__':
    parser = argparse.ArgumentParser()
    parser.add_argument("listen_proto",
                        help="Proto type to test http, https, ftp, \
                              swift, bzr+ssh or file")
    parser.add_argument("--port", type=int,
                        help="Port to listen on (http, https or ftp)")
    parser.add_argument("--swift-user", help="swift username")
    parser.add_argument("--swift-password", help="swift password")
    parser.add_argument("--swift-authurl", help="swift authurl")
    parser.add_argument("--swift-tenant", help="swift tenant")
    parser.add_argument('--anon', action='store_true', default=False,
                        help="Setup anonymous access (http, https or ftp)")
    parser.add_argument('--add-authkey', action='store_true', default=False,
                        help="Add generated public key to authorized_keys")

    args = parser.parse_args()

    default_ports = {'ftp': 2121, 'http': 8080, 'https': 4343}
    creds = {}
    if args.port:
        listen_port = args.port
    elif args.listen_proto in default_ports.keys():
        listen_port = default_ports[args.listen_proto]
    if not args.anon:
        password = ''.join(random.choice(string.ascii_letters +
                                         string.digits) for _ in range(8))
        creds = {'contentfetch': password}

    work_area = create_work_area()
    root = static.File(work_area)

    if args.listen_proto in ['http', 'https', 'ftp']:
        test_listener = setup_test_listener(work_area, listen_port,
                                            args.listen_proto, **creds)
        test_listener.display_access_info()
        test_listener.start_server()
    elif args.listen_proto == 'bzr+ssh':
        test_files = setup_test_files(work_area, args.listen_proto,
                                      add_pub_key=args.add_authkey)
        test_files.deploy_payload()
    elif args.listen_proto == 'swift':
        test_files = setup_test_files(work_area, args.listen_proto)
        if args.swift_user:
            test_files.set_swift_creds(args.swift_authurl, args.swift_user,
                                       args.swift_password, args.swift_tenant,
                                       "content.tar", "content-fetcher-test")
        elif 'OS_USERNAME' in os.environ:
            test_files.set_swift_creds(os.environ['OS_AUTH_URL'],
                                       os.environ['OS_USERNAME'],
                                       os.environ['OS_PASSWORD'],
                                       os.environ['OS_TENANT_NAME'],
                                       "content.tar",
                                       "content-fetcher-test")
        else:
            print "Could not discover swift creds"
            sys.exit(1)
        test_files.deploy_payload()
    elif args.listen_proto == 'file':
        test_files = setup_test_files(work_area, args.listen_proto)
        test_files.deploy_payload()
    else:
        print args.listen_proto + " test server not currently supported"
    sys.stdout.flush()
